import io
import json
import logging
import os
import platform
import subprocess
import sys
import tempfile

if __package__:
    from .general import populate_general_options
else:
    from scanners.general import populate_general_options


def populate_parser(args, parser):
    scanner = __name__.split(".")[-1]
    module_group = parser.add_argument_group(__name__, f"Scans with the {scanner} command-line scanner")
    populate_general_options(args, module_group)
    module_group.add_argument("--path", default=os.path.join(args.directory, "tools", scanner, "bin"),
                              help="path to scanner binary")


def run(args, target):
    if os.path.isdir(args.path):
        path = os.path.abspath(args.path)
    else:
        path = ""

    scanner = __name__.split(".")[-1]
    report = os.path.join(tempfile.gettempdir(), target.get_hash())

    scan_cmd = [os.path.join(path, scanner)]
    if platform.system() == "Windows":
        scan_cmd[0] = f"{scan_cmd[0]}.bat"
    scan_cmd += ["--report-save-path", f"{report}.bin"]
    scan_cmd += ["--output-only-positives"]
    scan_cmd += ["--scope-page-limit", "1"]
    scan_cmd += ["--scope-include-pattern", target.url.split("?", 1)[0]]
    if args.args:
        scan_cmd += args.args.split()
    scan_cmd += [target.url]

    report_cmd = [os.path.join(path, f"{scanner}_reporter")]
    if platform.system() == "Windows":
        report_cmd[0] = f"{report_cmd[0]}.bat"
    report_cmd += ["--reporter", f"json:outfile={report}.json"]
    report_cmd += [f"{report}.bin"]

    try:
        subprocess.run(scan_cmd, check=True, capture_output=True)
        subprocess.run(report_cmd, check=True, capture_output=True)
    except OSError as e:
        if "No such file or directory" in str(e) or "The system cannot find the file specified" in str(e):
            logging.critical(
                f"Could not find {scanner}. If not in PATH, extract or symlink as [directory]/tools/{scanner} "
                "or set path option to correct directory.")
            sys.exit(1)
        else:
            raise
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to execute command - {str(e)}\n{e.stderr.decode()}")
        return False

    with io.open(report + ".json", encoding="utf-8") as data_file:
        contents = data_file.read()
        data = json.loads(contents)
        vulns = []
        for issue in data["issues"]:
            vuln = {}
            vuln["vulnerability"] = issue["check"]["shortname"]
            vuln["url"] = issue["vector"]["url"]
            vuln["parameter"] = issue["vector"]["affected_input_name"]
            if vuln["parameter"] is None:
                vuln["parameter"] = ""
            vuln["method"] = issue["request"]["method"]
            vuln["poc"] = issue["response"]["url"]
            vuln["poc_data"] = issue["request"]["effective_body"]
            if vuln["poc_data"] is None:
                vuln["poc_data"] = ""
            vulns.append(vuln)

    os.remove(report + ".bin")
    os.remove(report + ".json")

    return vulns



import datetime
import hashlib
import importlib
import importlib.util
import ipaddress
import json
import logging
import os
import socket
from urllib.parse import parse_qsl, quote, urlencode, urlparse


def generate_fingerprint(url):
    url_parts = urlparse(url)
    netloc = url_parts.netloc
    depth = str(url_parts.path.count("/"))
    page = url_parts.path.split("/")[-1]
    params = []
    for param in url_parts.query.split("&"):
        split = param.split("=", 1)
        if len(split) == 2 and split[1]:
            params.append(split[0])
    fingerprint = "|".join((netloc, depth, page, ",".join(sorted(params))))
    return generate_hash(fingerprint)


def generate_timestamp():
    return datetime.datetime.now().astimezone().isoformat()


def generate_hash(url):
    return hashlib.md5(url.encode("utf-8")).hexdigest()


def parse_host(url):
    host = None
    try:
        host = urlparse(url).hostname
    except Exception:
        logging.debug(f"Failed to parse host from url: {url}")
        raise
    return host


def resolve_ip(host):
    ip = None
    try:
        ip = ipaddress.ip_address(socket.gethostbyname(host))
    except Exception:
        logging.debug(f"Failed to resolve ip address for host: {host}")
        raise
    return ip


def get_parsed_url(url):
    url_parts = urlparse(url)
    quoted_path = quote(url_parts.path)
    encoded_query = urlencode(parse_qsl(url_parts.query, keep_blank_values=True))
    parsed_url = url_parts._replace(path=quoted_path, query=encoded_query)
    return parsed_url.geturl()


def get_database_module(address):
    module_name = None

    if address.startswith("postgresql://"):
        for module in ["psycopg", "psycopg2"]:
            module_spec = importlib.util.find_spec(module)
            if module_spec:
                module_name = module
                break
        if not module_name:
            logging.error("Missing postgresql module - try: pip install psycopg[binary]")
            raise

    elif address.startswith("sqlite3://"):
        module = "sqlite3"
        module_spec = importlib.util.find_spec(module)
        if module_spec:
            module_name = module
        else:
            logging.error("Missing sqlite3 module - try: pip install sqlite3")
            raise

    else:
        logging.error(f"Unknown database protocol for address: {address}")
        raise ImportError

    return importlib.import_module(module_name, package=None)


def generate_report(url, start_time, end_time, label, vulnerabilities):
    vulns = {}
    vulns["vulnerabilities"] = vulnerabilities
    vulns["starttime"] = start_time
    vulns["endtime"] = end_time
    vulns["url"] = url
    vulns["label"] = label
    return vulns


def write_report(report, scanner_args, hash=None):
    if scanner_args.report_filename:
        report_filename = scanner_args.report_filename
    else:
        if not hash:
            hash = generate_hash(report["url"])
        report_filename = hash + ".json"

    filename = os.path.join(scanner_args.report_dir, report_filename)

    if scanner_args.report_append:
        report_mode = "a"
    else:
        report_mode = "w"

    indent = scanner_args.report_indent
    if indent and indent.isdigit():
        indent = int(indent)

    try:
        os.makedirs(os.path.abspath(scanner_args.report_dir), exist_ok=True)
        with open(filename, report_mode) as outfile:
            json.dump(report, outfile, indent=indent, sort_keys=True)
            outfile.write('\n')
            logging.info("Report saved to: %s" % outfile.name)
    except OSError as e:
        logging.error(f"Failed to write report - {str(e)}")
        raise



import json
import logging
import re
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from http.client import IncompleteRead
from itertools import repeat
from urllib.error import HTTPError, URLError
from urllib.parse import urlencode, urlparse
from urllib.request import urlopen

if __package__:
    from .general import populate_general_options
else:
    from indexers.general import populate_general_options


def populate_pywb_options(module_group):
    module_group.add_argument("--server", required=True,
                              help="pywb server url")
    module_group.add_argument("--domain", required=True,
                              help="pull all results for given domain or subdomain")
    module_group.add_argument("--cdx-api-suffix", default="/cdx",
                              help="suffix after index for index api")
    module_group.add_argument("--index",
                              help="search a specific index")
    module_group.add_argument("--field", default="url",
                              help="field (fl) to query")
    module_group.add_argument("--filter",
                              help="query filter to apply to the search")
    module_group.add_argument("--page-size", type=int,
                              help="number of results to request per page")


def populate_parser(_, parser):
    module_group = parser.add_argument_group(__name__, "Searches a given pywb server's crawl data")
    populate_general_options(module_group)
    populate_pywb_options(module_group)


def run(args):
    source = __name__.split(".")[-1]
    results, source = run_pywb(args, source)
    return results, source


def run_pywb(args, source, data={}):
    data["url"] = f"*.{args.domain}"
    data["output"] = "json"
    if args.filter:
        data["filter"] = args.filter

    if args.index is None:
        args.index = get_latest_index(args)

    if args.index:
        source += f",index:{args.index}"
    else:
        args.index = ""

    num_pages = get_num_pages(args, data)

    results = get_results(args, data, num_pages)
    for result in results:
        logging.debug(result)
    logging.info("Fetched %d results", len(results))
    return results, source


def issue_request(args, url):
    response = ""
    for i in range(args.retries + 1):
        try:
            logging.debug(url)
            response = urlopen(url).read().decode("utf-8")
        except (HTTPError, IncompleteRead, URLError) as e:
            if type(e).__name__ == "HTTPError" and e.code == 404:
                response_str = e.read().decode("utf-8")
                if "message" in response_str:
                    message = json.loads(response_str)["message"]
                else:
                    message = str(e)
                logging.error(f"Request failed - {message}")
                raise
            elif i == args.retries:
                logging.error(f"Request failed - {str(e)}")
                raise
            else:
                logging.debug(f"Request failed (attempt {i + 1} of {args.retries}) - {str(e)}")
                time.sleep(2**i)
                continue
        break

    return response


def get_latest_index(args):
    logging.debug("Fetching latest index list")
    url = f"{args.server}/collinfo.json"

    try:
        response_str = issue_request(args, url)
    except Exception:
        sys.exit(1)

    try:
        response = json.loads(response_str)
    except json.decoder.JSONDecodeError:
        logging.error(f"Unexpected response:\n{response_str}")
        sys.exit(1)

    try:
        if "fixed" in response:
            fixed = response["fixed"]
            dynamic = response["dynamic"]
            index = sorted(fixed)[-1] if fixed else sorted(dynamic)[-1]
        else:
            index = response[0]["id"]
    except Exception:
        logging.error(f"Unexpected response:\n{response}")
        sys.exit(1)
    return index


def get_num_pages(args, data):
    logging.debug("Fetching number of pages")
    data["showNumPages"] = "true"
    if args.page_size:
        data["pageSize"] = args.page_size
    url = f"{args.server}/{args.index}{args.cdx_api_suffix}?{urlencode(data)}"

    try:
        response_str = issue_request(args, url)
    except Exception:
        sys.exit(1)

    try:
        response = json.loads(response_str)
    except json.decoder.JSONDecodeError:
        logging.error(f"Unexpected response:\n{response_str}")
        sys.exit(1)

    if "pages" in response:
        num_pages = int(response["pages"])
    elif response[0] and response[0][0] == "numpages":
        num_pages = int(response[1][0])
    else:
        logging.error(f"Unexpected response:\n{response}")
        sys.exit(1)

    del data["showNumPages"]
    logging.debug("Got %d pages", num_pages)
    return num_pages


def get_page(args, data, page):
    logging.debug("Fetching page %d", page)
    data["fl"] = args.field
    data["page"] = page
    url = f"{args.server}/{args.index}{args.cdx_api_suffix}?{urlencode(data)}"

    response_str = issue_request(args, url)

    if response_str.startswith("["):
        try:
            response = json.loads(response_str.replace("\n", ""))[1:]
        except json.decoder.JSONDecodeError:
            logging.error(f"Unexpected response:\n{response_str}")
            sys.exit(1)
    else:
        response = response_str.splitlines()

    pattern = r"http[s]?://([^/.]*\.)*" + args.domain + "(/|$)"
    domain_url = re.compile(pattern)

    results = set()
    for item in response:
        if isinstance(item, list):
            item_url = item[0]
        else:
            try:
                item_url = json.loads(item)["url"]
            except json.decoder.JSONDecodeError:
                logging.error(f"Unexpected item:\n{item}")
                break

        parsed_url = urlparse(item_url.strip())
        url = parsed_url._replace(netloc=parsed_url.hostname).geturl()

        if domain_url.match(url):
            results.add(url)
    return results


def get_results(args, data, num_pages):
    results = set()
    try:
        if args.threads == 1:
            for page in range(num_pages):
                results.update(get_page(args, data, page))
        else:
            with ThreadPoolExecutor(max_workers=args.threads) as executor:
                threads = executor.map(get_page, repeat(args), repeat(data), range(num_pages))
                for result in threads:
                    results.update(result)
    except Exception:
        logging.error("Failed to fetch all results")
        pass

    return list(results)



import io
import json
import logging
import os
import platform
import subprocess
import sys
import tempfile
from urllib.parse import urlparse, urlunparse, urljoin

if __package__:
    from .general import populate_general_options
else:
    from scanners.general import populate_general_options


def populate_parser(args, parser):
    scanner = __name__.split(".")[-1]
    module_group = parser.add_argument_group(__name__, f"Scans with the {scanner} command-line scanner")
    populate_general_options(args, module_group)
    module_group.add_argument("--path", default=os.path.join(args.directory, "tools", scanner, "bin"),
                              help="path to scanner binary")


def run(args, target):
    if os.path.isdir(args.path):
        path = os.path.abspath(args.path)
    else:
        path = ""

    scanner = __name__.split(".")[-1]
    report = os.path.join(tempfile.gettempdir(), target.get_hash())

    scan_cmd = [os.path.join(path, scanner)]
    if platform.system() == "Windows" and path != "":
        scan_cmd.insert(0, sys.executable)
    scan_cmd += ["--url", target.url]
    scan_cmd += ["--scope", "page"]
    scan_cmd += ["--flush-session"]
    scan_cmd += ["--format", "json"]
    scan_cmd += ["--output", f"{report}.json"]
    if args.args:
        scan_cmd += args.args.split()

    try:
        subprocess.run(scan_cmd, check=True, capture_output=True)
    except OSError as e:
        if "No such file or directory" in str(e) or "The system cannot find the file specified" in str(e):
            logging.critical(
                f"Could not find {scanner}. If not in PATH, extract or symlink as [directory]/tools/{scanner}"
                "or set path option to correct directory.")
            sys.exit(1)
        else:
            raise
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to execute command - {str(e)}\n{e.stderr.decode()}")
        return False

    with io.open(report + ".json", encoding="utf-8") as data_file:
        contents = data_file.read()
        data = json.loads(contents)
        vulns = []
        for vuln_type in data["vulnerabilities"]:
            for vulnerability in data["vulnerabilities"][vuln_type]:
                url = urlparse(data["infos"]["target"])
                poc_request = vulnerability["http_request"].split("\n")
                poc_path = poc_request[0].split(" ")[1]

                vuln = {}
                vuln["vulnerability"] = vuln_type
                vuln["url"] = urlunparse(url)
                vuln["parameter"] = vulnerability["parameter"]
                vuln["method"] = vulnerability["method"]
                vuln["poc"] = urljoin(vuln["url"], poc_path)
                if vuln["method"] == "POST":
                    vuln["poc_data"] = poc_request[-1]
                else:
                    vuln["poc_data"] = ""

                vulns.append(vuln)

    os.remove(report + ".json")

    return vulns



#!/usr/bin/env python3
if __package__:
    from dorkbot.util import get_database_module
else:
    from util import get_database_module
import logging
import os
import time
from contextlib import closing


class Database:
    def __init__(self, address, retries, retry_on):
        self.address = address
        self.retries = retries
        self.retry_on = retry_on

        if address.startswith("postgresql://"):
            self.module = get_database_module(address)
            self.database = address
            self.id_type = "INTEGER GENERATED ALWAYS AS IDENTITY PRIMARY KEY"
            self.insert = "INSERT"
            self.conflict = "ON CONFLICT DO NOTHING"
            self.param = "%s"
            self.connect_kwargs = {}

        elif address.startswith("sqlite3://"):
            self.module = get_database_module(address)
            self.database = os.path.expanduser(address[10:])
            self.id_type = "INTEGER PRIMARY KEY"
            self.insert = "INSERT OR REPLACE"
            self.conflict = ""
            self.param = "?"
            self.connect_kwargs = {}

        else:
            self.database = None

    def connect(self):
        for i in range(self.retries + 1):
            try:
                self.db = self.module.connect(self.database, **self.connect_kwargs)
                break
            except self.module.Error as e:
                if i < self.retries and any(string in str(e) for string in self.retry_on):
                    logging.warning(f"Database connection failed (attempt {i + 1} of {self.retries}) - {str(e)}")
                    time.sleep(2 ** (5 + i))
                    continue
                else:
                    logging.error(f"Database connection failed - {str(e)}")
                    raise

    def close(self):
        self.db.close()

    def execute(self, *sql, fetch=0):
        statement, parameters = (sql[0], sql[1] if len(sql) == 2 else ())

        for i in range(self.retries + 1):
            try:
                with closing(self.db.cursor()) as c:
                    result = None
                    if isinstance(parameters, list):
                        c.executemany(statement, parameters)
                    else:
                        c.execute(statement, parameters)
                    if fetch is True:
                        result = c.fetchall()
                    elif fetch == 1:
                        result = c.fetchone()
                    elif fetch > 1:
                        result = c.fetchmany(fetch)
                self.db.commit()
                return result
            except self.module.Error as e:
                if i < self.retries and any(error in str(e) for error in self.retry_on):
                    logging.warning(f"Database execution failed (attempt {i + 1} of {self.retries}) - {str(e)}")
                    self.close()
                    time.sleep(2 ** (5 + i))
                    self.connect()
                    continue
                else:
                    logging.error(f"Database execution failed - {str(e)}")
                    raise



#!/usr/bin/env python3
if __package__:
    from dorkbot.util import generate_hash, parse_host, resolve_ip
else:
    from util import generate_hash, parse_host, resolve_ip


class Target:
    def __init__(self, url):
        self.url = url
        self._host = None
        self._ip = None
        self._hash = None

    def get_host(self):
        if not self._host:
            self._host = parse_host(self.url)
        return self._host

    def get_ip(self):
        if not self._ip:
            host = self.get_host()
            self._ip = resolve_ip(host)

    def get_hash(self):
        if not self._hash:
            self._hash = generate_hash(self.url)
        return self._hash



from urllib.parse import urlparse


def populate_parser(args, parser):
    module_group = parser.add_argument_group(__name__, "Example module that returns a few urls")  # noqa:F841


def run(args):
    source = __name__.split(".")[-1]
    urls = [
        "http://www.example.com/foo.php?id=4",
        "http://admin.example.com/bar.php?page=home",
        "http://www.example.com/baz.jsp?size=124"
    ]

    return [urlparse(item.strip()).geturl() for item in urls], source



import json
import logging
import sys
import time
from urllib.error import HTTPError
from urllib.parse import urlencode, urlparse
from urllib.request import urlopen


def populate_parser(args, parser):
    module_group = parser.add_argument_group(__name__, "Searches google.com")
    module_group.add_argument("--key", required=True,
                              help="API key")
    module_group.add_argument("--engine", required=True,
                              help="CSE id")
    module_group.add_argument("--query", required=True,
                              help="search query")
    module_group.add_argument("--domain",
                              help="limit searches to specified domain")


def run(args):
    source = __name__.split(".")[-1]
    results = get_results(args.key, args.engine, args.query, args.domain)
    return results, source


def get_results(key, engine, query, domain):
    data = {}
    data["key"] = key
    data["cx"] = engine
    data["q"] = query
    data["num"] = 10
    data["start"] = 0

    if domain:
        data["siteSearch"] = domain

    results = []
    while True:
        items = issue_request(data)
        data["start"] += data["num"]
        if not items:
            break
        results.extend(items)

    for result in results:
        logging.debug(result)
    logging.info("Fetched %d results", len(results))
    return results


def issue_request(data):
    url = "https://www.googleapis.com/customsearch/v1?" + urlencode(data)
    while True:
        try:
            logging.debug("Issuing request: %s", url)
            response_str = urlopen(url)
            response_str = response_str.read().decode("utf-8")
            response = json.loads(response_str)
            break
        except HTTPError as e:
            response_str = e.read().decode("utf-8")
            response = json.loads(response_str)
            if "Invalid Value" in response["error"]["message"]:
                return []
            elif "Request contains an invalid argument" in response["error"]["message"]:
                return []
            for error in response["error"]["errors"]:
                logging.error("%s::%s::%s", error["domain"], error["reason"], error["message"])
            if "User Rate Limit Exceeded" in response["error"]["message"]:
                time.sleep(5)
                continue
            elif "Daily Limit Exceeded" in response["error"]["message"]:
                logging.info("sleeping 1 hour")
                time.sleep(3600)
                continue
            else:
                logging.error("Failed to fetch results - %d %s", response["error"]["code"],
                              response["error"]["message"])
                sys.exit(1)

    items = []
    # https://developers.google.com/custom-search/v1/reference/rest/v1/Search
    if int(response['searchInformation'].get('totalResults', 0)) == 0:
        return []
    for item in response["items"]:
        items.append(urlparse(item["link"]).geturl())

    return items



#!/usr/bin/env python3
import ipaddress
import logging
import os
import re
if __package__:
    from dorkbot.database import Database
else:
    from database import Database


class Blocklist(Database):
    def __init__(self, address, drop_tables=False, create_tables=False, retries=0, retry_on=[]):
        Database.__init__(self, address, retries, retry_on)
        self.ip_set = set()
        self.host_set = set()
        self.regex_set = set()

        if self.database:
            if address.startswith("sqlite3://"):
                try:
                    os.makedirs(os.path.dirname(os.path.abspath(self.database)), exist_ok=True)
                except OSError as e:
                    logging.error(f"Failed to create parent directory for database file - {str(e)}")
                    raise

            self.connect()

            if drop_tables:
                logging.debug("Dropping tables")
                self.execute("DROP TABLE IF EXISTS blocklist")

            if create_tables:
                self.execute("CREATE TABLE IF NOT EXISTS blocklist"
                             f" (id {self.id_type},"
                             " item VARCHAR)")

        else:
            try:
                os.makedirs(os.path.dirname(os.path.abspath(self.address)), exist_ok=True)
                self.blocklist_file = open(self.address, "r")
            except OSError as e:
                logging.error(f"Failed to open database file - {str(e)}")
                raise

        self.parse_list(self.read_items())

    def connect(self):
        if self.database:
            Database.connect(self)
        else:
            try:
                self.blocklist_file = open(self.address, "a")
            except Exception as e:
                logging.error(f"Failed to read blocklist file - {str(e)}")
                raise

    def close(self):
        if self.database:
            self.db.close()
        else:
            self.address.close()

    def parse_list(self, items):
        for item in items:
            if item.startswith("ip:"):
                ip = item.split(":")[1]
                try:
                    ip_net = ipaddress.ip_network(ip)
                except ValueError as e:
                    logging.error(f"Could not parse blocklist item as ip - {str(e)}")
                    continue
                self.ip_set.add(ip_net)
            elif item.startswith("host:"):
                self.host_set.add(item.split(":")[1])
            elif item.startswith("regex:"):
                self.regex_set.add(item.split(":")[1])
            else:
                logging.warning(f"Could not parse blocklist item - {item}")

        pattern = "|".join(self.regex_set)
        if pattern:
            self.regex = re.compile(pattern)
        else:
            self.regex = None

    def get_parsed_items(self):
        parsed_ip_set = set()
        for ip_net in self.ip_set:
            if ip_net.num_addresses == 1:
                parsed_ip_set.add(str(ip_net[0]))
            else:
                parsed_ip_set.add(str(ip_net))

        return ["ip:" + item for item in parsed_ip_set] + \
               ["host:" + item for item in self.host_set] + \
               ["regex:" + item for item in self.regex_set]

    def read_items(self):
        if self.database:
            rows = self.execute("SELECT item FROM blocklist ORDER BY id ASC", fetch=True)
            items = [row[0] for row in rows] if rows else []
        else:
            items = self.blocklist_file.read().splitlines()

        return items

    def add(self, item):
        self.connect()

        if item.startswith("ip:"):
            ip = item.split(":")[1]
            try:
                ip_net = ipaddress.ip_network(ip)
            except ValueError as e:
                logging.error(f"Could not parse blocklist item as ip - {str(e)}")
                raise
            self.ip_set.add(ip_net)
        elif item.startswith("host:"):
            self.host_set.add(item.split(":")[1])
        elif item.startswith("regex:"):
            self.regex_set.add(item.split(":")[1])
        else:
            logging.error("Could not parse blocklist item - %s", item)
            raise

        self.execute("%s INTO blocklist (item) VALUES (%s)" % (self.insert, self.param), (item,))
        self.close()

    def delete(self, item):
        self.connect()
        self.execute("DELETE FROM blocklist WHERE item=(%s)" % self.param, (item,))
        self.close()

    def match(self, target):
        if self.regex and self.regex.match(target.url):
            return True

        if target.get_host() in self.host_set:
            return True

        for ip_net in self.ip_set:
            if target.get_ip() in ip_net:
                return True

        return False

    def flush(self):
        logging.info("Flushing blocklist")
        if self.database:
            self.execute("DELETE FROM blocklist")
        else:
            try:
                os.unlink(self.address)
            except OSError as e:
                logging.error(f"Failed to delete blocklist file - {str(e)}")
                raise

        self.regex = None
        self.regex_set = set()
        self.ip_set = set()
        self.host_set = set()



import argparse

if __package__:
    from .pywb import populate_pywb_options, run_pywb
    from .general import populate_general_options
else:
    from indexers.pywb import populate_pywb_options, run_pywb
    from indexers.general import populate_general_options


def populate_parser(_, parser):
    module_group = parser.add_argument_group(__name__, "Searches commoncrawl.org crawl data")
    populate_general_options(module_group)
    populate_pywb_options(module_group)

    defaults = {"server": "https://index.commoncrawl.org", "cdx_api_suffix": "-index", "field": "url"}
    module_group.set_defaults(**defaults)
    for action in module_group._actions:
        if action.dest in defaults.keys():
            action.help = argparse.SUPPRESS
            action.required = False


def run(args):
    source = __name__.split(".")[-1]
    results, source = run_pywb(args, source)
    return results, source



import io
import logging
import sys
from urllib.parse import urlparse


def populate_parser(args, parser):
    module_group = parser.add_argument_group(__name__, "Accepts urls from stdin, one per line")  # noqa:F841


def run(args):
    source = __name__.split(".")[-1]
    with io.open(sys.stdin.fileno(), encoding="utf-8") as stdin:
        results = [urlparse(item.strip()).geturl() for item in stdin]

        for result in results:
            logging.debug(result)
        logging.info("Fetched %d results", len(results))
        return results, source



#!/usr/bin/env python3
if __package__:
    from dorkbot.database import Database
    from dorkbot.target import Target
    from dorkbot.util import generate_fingerprint, get_parsed_url
else:
    from database import Database
    from target import Target
    from util import generate_fingerprint, get_parsed_url
import logging
import os


class TargetDatabase(Database):
    def __init__(self, address, drop_tables=False, create_tables=False, retries=0, retry_on=[]):
        protocols = ["postgresql://", "sqlite3://"]
        if not any(address.startswith(protocol) for protocol in protocols):
            address = f"sqlite3://{address}"
        Database.__init__(self, address, retries, retry_on)

        if self.database and address.startswith("sqlite3://"):
            try:
                os.makedirs(os.path.dirname(os.path.abspath(self.database)), exist_ok=True)
            except OSError as e:
                logging.error(f"Failed to create parent directory for database file - {str(e)}")
                raise

        self.connect()

        if drop_tables:
            logging.debug("Dropping tables")
            self.execute("DROP TABLE IF EXISTS targets")
            self.execute("DROP TABLE IF EXISTS sources")
            self.execute("DROP TABLE IF EXISTS fingerprints")
            self.execute("DROP TABLE IF EXISTS blocklist")

        if create_tables:
            self.execute("CREATE TABLE IF NOT EXISTS targets"
                         f" (id {self.id_type},"
                         " url VARCHAR UNIQUE,"
                         " source_id INTEGER,"
                         " fingerprint_id INTEGER,"
                         " scanned INTEGER DEFAULT 0)")
            self.execute("CREATE TABLE IF NOT EXISTS sources"
                         f" (id {self.id_type},"
                         " source VARCHAR UNIQUE)")
            self.execute("CREATE TABLE IF NOT EXISTS fingerprints"
                         f" (id {self.id_type},"
                         " fingerprint VARCHAR UNIQUE,"
                         " scanned INTEGER DEFAULT 0)")
            self.execute("CREATE TABLE IF NOT EXISTS blocklist"
                         f" (id {self.id_type},"
                         " item VARCHAR UNIQUE)")

    def get_urls(self, args):
        options = {"unscanned_only": args.unscanned_only, "count": args.count}
        sql, parameters = self.get_targets_query(args, **options)
        rows = self.execute(sql, parameters, fetch=True) or []
        urls = []
        for row in rows:
            url = row[0]
            if args.source is True:
                url += f" | {row[-1]}"
            urls.append(url)
        return urls

    def get_targets_query(self, args, unscanned_only=False, count=-1):
        fields = ["t.url", "t.id", "t.fingerprint_id", "f.fingerprint"]
        join = [("LEFT", "fingerprints f ON f.id = t.fingerprint_id")]
        where = []
        parameters = ()

        if unscanned_only:
            where.extend(["t.scanned = '0'", "(t.fingerprint_id IS NULL OR f.scanned = '0')"])

        if args.source:
            fields.append("s.source")
            if args.source is True:
                join_type = "LEFT"
            else:
                join_type = "INNER"
                where.append("s.source = %s" % self.param)
                parameters = (args.source,)
            join.append((join_type, "sources s ON s.id = t.source_id"))

        sql = "SELECT "
        for i, field in enumerate(fields):
            sql += "" if i == 0 else ", "
            sql += field
        sql += " FROM targets t"
        for join_type, join_criteria in join:
            sql += f" {join_type} JOIN {join_criteria}"
        for i, clause in enumerate(where):
            sql += " WHERE " if i == 0 else " AND "
            sql += clause
        sql += " ORDER BY "
        sql += "RANDOM()" if args.random else "t.id ASC"
        sql += f" LIMIT {count}" if count > 0 else ""

        return sql, parameters

    def matches_blocklists(self, url, blocklists):
        target = Target(url)
        for blocklist in blocklists:
            if blocklist.match(target):
                return True
        return False

    def get_next_target(self, args, blocklists=[]):
        options = {"unscanned_only": True}
        sql, parameters = self.get_targets_query(args, **options)
        target = None
        fingerprints = {}
        while True:
            row = self.execute(sql, parameters, fetch=1)
            if not row:
                break
            url, target_id, fingerprint_id, fingerprint, *_ = row

            try:
                if self.matches_blocklists(url, blocklists):
                    logging.debug(f"Deleting (matches blocklist pattern): {url}")
                    self.delete_target(url)
                    continue
            except Exception:
                if args.delete_on_error:
                    self.delete_target(url)
                    continue

            if fingerprint_id:
                logging.debug(f"Found unique fingerprint: {url}")
                if not args.test:
                    self.mark_fingerprint_scanned(fingerprint_id)
                target = url

            else:
                logging.debug(f"Computing fingerprint: {url}")
                fingerprint = generate_fingerprint(url)

                if fingerprint in fingerprints:
                    logging.debug(f"Skipping (matches existing fingerprint): {url}")
                    fingerprint_id = fingerprints[fingerprint]
                else:
                    fingerprint_id = self.get_fingerprint_id(fingerprint)
                    if fingerprint_id:
                        logging.debug(f"Skipping (matches scanned fingerprint): {url}")
                        fingerprints[fingerprint] = fingerprint_id
                    else:
                        logging.debug(f"Found unique fingerprint: {url}")
                        fingerprint_id = self.add_fingerprint(fingerprint, scanned=(not args.test))
                        target = url
                self.update_target_fingerprint(target_id, fingerprint_id)

            if target:
                break
        return target

    def add_target(self, url, source=None, blocklists=[]):
        try:
            if self.matches_blocklists(url, blocklists):
                logging.debug(f"Ignoring (matches blocklist pattern): {url}")
                return
        except Exception:
            pass

        logging.debug(f"Adding target {url}")
        if source:
            source_id = self.get_source_id(source)
            if not source_id:
                source_id = self.add_source(source)
        else:
            source_id = None

        self.execute("%s INTO targets (url, source_id) VALUES (%s, %s) %s"
                     % (self.insert, self.param, self.param, self.conflict),
                     (get_parsed_url(url), source_id))

    def add_targets(self, urls, source=None, blocklists=[], chunk_size=1000):
        logging.debug(f"Adding {len(urls)} targets")
        if source:
            source_id = self.get_source_id(source)
            if not source_id:
                source_id = self.add_source(source)
        else:
            source_id = None

        for x in range(0, len(urls), chunk_size):
            urls_chunk = urls[x:x + chunk_size]
            urls_chunk_add = []
            for url in urls_chunk:
                try:
                    if self.matches_blocklists(url, blocklists):
                        logging.debug(f"Ignoring (matches blocklist pattern): {url}")
                        continue
                except Exception:
                    pass
                urls_chunk_add.append(get_parsed_url(url))

            self.execute("%s INTO targets (url, source_id) VALUES (%s, %s) %s"
                         % (self.insert, self.param, self.param, self.conflict),
                         [(url, source_id) for url in urls_chunk_add])

    def mark_target_scanned(self, target_id):
        self.execute("UPDATE targets SET scanned = 1 WHERE id = %s" % self.param, (target_id,))

    def delete_target(self, url):
        logging.debug(f"Deleting target {url}")
        self.execute("DELETE FROM targets WHERE url = %s" % self.param, (url,))

    def flush_targets(self):
        logging.info("Flushing targets")
        self.execute("DELETE FROM targets")
        self.execute("DELETE FROM sources")

    def add_source(self, source):
        logging.debug(f"Adding source {source}")
        row = self.execute("%s INTO sources (source) VALUES (%s) %s RETURNING id"
                           % (self.insert, self.param, self.conflict),
                           (source,), fetch=1)
        return row if not row else row[0]

    def get_source_id(self, source):
        row = self.execute("SELECT id FROM sources WHERE source = %s"
                           % self.param, (source,), fetch=1)
        return row if not row else row[0]

    def get_sources(self):
        rows = self.execute("SELECT source FROM sources ORDER BY id ASC", fetch=True)
        return [row[0] for row in rows] if rows else []

    def add_fingerprint(self, fingerprint, scanned=False):
        logging.debug(f"Adding fingerprint {fingerprint}")
        row = self.execute("%s INTO fingerprints (fingerprint, scanned) VALUES (%s, %s) %s RETURNING id"
                           % (self.insert, self.param, self.param, self.conflict),
                           (fingerprint, 1 if scanned else 0), fetch=1)
        return row if not row else row[0]

    def update_target_fingerprint(self, target_id, fingerprint_id):
        logging.debug(f"Updating target fingerprint id {target_id}->{fingerprint_id}")
        self.execute("UPDATE targets SET fingerprint_id = %s WHERE id = %s"
                     % (self.param, self.param), (fingerprint_id, target_id))

    def flush_fingerprints(self):
        logging.info("Flushing fingerprints")
        self.execute("UPDATE targets SET fingerprint_id = NULL")
        self.execute("DELETE FROM fingerprints")

    def reset_scanned(self):
        logging.info("Resetting scanned")
        self.execute("UPDATE targets SET scanned = 0")
        self.execute("UPDATE fingerprints SET scanned = 0")

    def get_fingerprint_id(self, fingerprint):
        row = self.execute("SELECT id FROM fingerprints WHERE fingerprint = %s"
                           % self.param, (fingerprint,), fetch=1)
        return row if not row else row[0]

    def mark_fingerprint_scanned(self, fingerprint_id):
        self.execute("UPDATE fingerprints SET scanned = 1 WHERE id = %s" % self.param, (fingerprint_id,))

    def prune(self, blocklists, args):
        logging.info("Pruning database")
        options = {"unscanned_only": args.unscanned_only, "count": args.count}
        sql, parameters = self.get_targets_query(args, **options)
        targets = self.execute(sql, parameters, fetch=True)
        if not targets:
            return
        targets.reverse()
        fingerprints = {}
        while targets:
            url, target_id, fingerprint_id, fingerprint, *_ = targets.pop()

            try:
                if self.matches_blocklists(url, blocklists):
                    logging.debug(f"Deleting (matches blocklist pattern): {url}")
                    self.delete_target(url)
            except Exception:
                if args.delete_on_error:
                    self.delete_target(url)
                    continue

            if fingerprint_id:
                if fingerprint in fingerprints:
                    fingerprint_id, fingerprint_count = fingerprints[fingerprint]
                    if fingerprint_count >= args.fingerprint_max:
                        logging.debug(f"Deleting (exceeds max fingerprint count): {url}")
                        self.delete_target(url)
                    else:
                        logging.debug(f"Skipping (matches existing fingerprint): {url}")
                        fingerprints[fingerprint] = (fingerprint_id, fingerprint_count + 1)
                        self.mark_target_scanned(target_id)
                else:
                    logging.debug(f"Found unique fingerprint: {url}")
                    fingerprints[fingerprint] = (fingerprint_id, 1)

            else:
                logging.debug(f"Computing fingerprint: {url}")
                fingerprint = generate_fingerprint(url)

                if fingerprint in fingerprints:
                    fingerprint_id, fingerprint_count = fingerprints[fingerprint]
                    if fingerprint_count >= args.fingerprint_max:
                        logging.debug(f"Deleting (exceeds max fingerprint count): {url}")
                        self.delete_target(url)
                    else:
                        logging.debug(f"Skipping (matches existing fingerprint): {url}")
                        fingerprints[fingerprint] = (fingerprint_id, fingerprint_count + 1)
                        self.mark_target_scanned(target_id)
                else:
                    fingerprint_id = self.get_fingerprint_id(fingerprint)
                    if fingerprint_id:
                        logging.debug(f"Skipping (matches existing fingerprint): {url}")
                    else:
                        logging.debug(f"Found unique fingerprint: {url}")
                        fingerprint_id = self.add_fingerprint(fingerprint, scanned=False)
                    fingerprints[fingerprint] = (fingerprint_id, 1)

                self.update_target_fingerprint(target_id, fingerprint_id)

    def get_fingerprintless_query(self, args):
        join, where = [], []

        if args.source and args.source is not True:
            join.append(("INNER", "sources s ON s.id = t.source_id"))
            where.append("s.source = %s" % self.param)
            parameters = (args.source,)
        else:
            parameters = ()

        sql = "SELECT t.url, t.id FROM targets t"
        for join_type, join_criteria in join:
            sql += f" {join_type} JOIN {join_criteria}"
        sql += " WHERE t.fingerprint_id IS NULL"
        for clause in where:
            sql += f" AND {clause}"
        sql += f" LIMIT {args.count}" if args.count > 0 else ""

        return sql, parameters

    def generate_fingerprints(self, args):
        logging.info("Generating fingerprints")
        sql, parameters = self.get_fingerprintless_query(args)
        targets = self.execute(sql, parameters, fetch=True)
        if targets:
            targets.reverse()
        fingerprints = {}
        while targets:
            url, target_id = targets.pop()
            fingerprint = generate_fingerprint(url)
            if fingerprint in fingerprints:
                fingerprint_id, fingerprint_count = fingerprints[fingerprint]
                if fingerprint_count >= args.fingerprint_max:
                    logging.debug(f"Deleting (exceeds max fingerprint count): {url}")
                    self.delete_target(url)
                    continue
                else:
                    fingerprints[fingerprint] = (fingerprint_id, fingerprint_count + 1)
            else:
                fingerprint_id = self.get_fingerprint_id(fingerprint)
                if fingerprint_id:
                    fingerprints[fingerprint] = (fingerprint_id, 1)
                else:
                    fingerprint_id = self.add_fingerprint(fingerprint, scanned=False)
                    fingerprints[fingerprint] = (fingerprint_id, 1)
            self.update_target_fingerprint(target_id, fingerprint_id)



import argparse

if __package__:
    from .pywb import populate_pywb_options, run_pywb
    from .general import populate_general_options
else:
    from indexers.pywb import populate_pywb_options, run_pywb
    from indexers.general import populate_general_options


def populate_parser(_, parser):
    module_group = parser.add_argument_group(__name__, "Searches archive.org crawl data")
    populate_general_options(module_group)
    populate_pywb_options(module_group)
    module_group.add_argument("--from", dest="from_", metavar="FROM",
                              help="beginning timestamp")
    module_group.add_argument("--to",
                              help="end timestamp")

    defaults = {"server": "https://web.archive.org", "cdx_api_suffix": "cdx/search/cdx", "field": "original", "index": False}
    module_group.set_defaults(**defaults)
    for action in module_group._actions:
        if action.dest in defaults.keys():
            action.help = argparse.SUPPRESS
            action.required = False


def run(args):
    source = __name__.split(".")[-1]
    data = {"collapse": "urlkey"}

    if args.from_:
        data["from"] = args.from_
        source += f",from:{args.from_}"
    if args.to:
        data["to"] = args.to
        source += f",to:{args.to}"

    results, source = run_pywb(args, source, data)
    return results, source



#!/usr/bin/env python3
if __package__:
    from ._version import __version__
    from dorkbot.target import Target
    from dorkbot.targetdatabase import TargetDatabase
    from dorkbot.blocklist import Blocklist
    from dorkbot.util import generate_timestamp, generate_report, write_report
else:
    from _version import __version__
    from target import Target
    from targetdatabase import TargetDatabase
    from blocklist import Blocklist
    from util import generate_timestamp, generate_report, write_report
import argparse
import configparser
import importlib
import logging
import os
import signal
import sys
from logging.handlers import WatchedFileHandler


def main():
    signal.signal(signal.SIGINT, graceful_shutdown)
    args, parser = get_main_args_parser()
    initialize_logger(args.log, args.verbose)

    if args.directory:
        try:
            os.makedirs(os.path.abspath(args.directory), exist_ok=True)
        except OSError as e:
            logging.error(f"Failed to create dorkbot directory - {str(e)}")
            sys.exit(1)

    if args.help:
        indexer_parser = None
        if args.indexer:
            indexer_parser, _ = get_module_parser(load_module("indexers", args.indexer))
            if not args.scanner:
                indexer_parser.print_help()
        if args.scanner:
            scanner_parser, _ = get_module_parser(load_module("scanners", args.scanner), parent_parser=indexer_parser)
            scanner_parser.print_help()
        if not args.indexer and not args.scanner:
            parser.print_help()
        sys.exit(0)

    if args.indexer or args.prune or args.scanner \
            or args.list_targets or args.flush_targets \
            or args.add_target or args.delete_target \
            or args.list_blocklist or args.flush_blocklist \
            or args.add_blocklist_item or args.delete_blocklist_item \
            or args.flush_fingerprints or args.generate_fingerprints \
            or args.reset_scanned or args.list_sources:

        retry = {"retries": args.retries, "retry_on": args.retry_on}

        try:
            tables = {"drop_tables": args.drop_tables, "create_tables": True}
            db = TargetDatabase(args.database, **tables, **retry)
            blocklist = Blocklist(db.address, **tables, **retry)
        except Exception:
            sys.exit(1)

        blocklists = [blocklist]
        if args.external_blocklist:
            for external_blocklist in args.external_blocklist:
                try:
                    blocklists.append(Blocklist(external_blocklist, **retry))
                except Exception:
                    sys.exit(1)

        if args.flush_blocklist:
            blocklist.flush()
        if args.add_blocklist_item:
            blocklist.add(args.add_blocklist_item)
        if args.delete_blocklist_item:
            blocklist.delete(args.delete_blocklist_item)
        if args.list_blocklist:
            for blocklist in blocklists:
                for item in blocklist.get_parsed_items():
                    print(item)

        if args.flush_fingerprints:
            db.flush_fingerprints()
        if args.reset_scanned:
            db.reset_scanned()
        if args.flush_targets:
            db.flush_targets()
        if args.add_target:
            db.add_target(args.add_target, source=args.source, blocklists=blocklists)
        if args.delete_target:
            db.delete_target(args.delete_target)

        if args.indexer:
            indexer_module = load_module("indexers", args.indexer)
            indexer_parser, _ = get_module_parser(indexer_module)
            indexer_args = indexer_parser.parse_args(format_module_args(args.indexer_arg))
            index(db, blocklists, indexer_module, args, indexer_args)

        if args.generate_fingerprints:
            db.generate_fingerprints(args)

        if args.prune:
            db.prune(blocklists, args)

        if args.scanner:
            scanner_module = load_module("scanners", args.scanner)
            scanner_parser, _ = get_module_parser(scanner_module)
            scanner_args = scanner_parser.parse_args(format_module_args(args.scanner_arg))
            scan(db, blocklists, scanner_module, args, scanner_args)

        if args.list_targets:
            try:
                urls = db.get_urls(args)
                for url in urls:
                    print(url)
            except BrokenPipeError:
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                sys.exit(1)

        if args.list_sources:
            sources = db.get_sources()
            for source in sources:
                print(source)

        db.close()
    else:
        parser.print_usage()

    logging.shutdown()


def graceful_shutdown(*_):
    sys.exit(0)


def initialize_logger(log_file, verbose):
    log = logging.getLogger()

    log_formatter = logging.Formatter(fmt="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z")
    if log_file:
        try:
            os.makedirs(os.path.dirname(os.path.abspath(log_file)), exist_ok=True)
            log_filehandler = WatchedFileHandler(log_file, mode="a", encoding="utf-8")
            log_filehandler.setLevel(logging.DEBUG)
            log_filehandler.setFormatter(log_formatter)
            log.addHandler(log_filehandler)
        except OSError as e:
            logging.error(f"Failed to create log file - {str(e)}")
            raise
    else:
        class LogFilter(logging.Filter):
            def __init__(self, level):
                self.level = level

            def filter(self, record):
                return record.levelno <= self.level

        log_stdouthandler = logging.StreamHandler(sys.stdout)
        log_stdouthandler.setLevel(logging.DEBUG)
        log_stdouthandler.addFilter(LogFilter(logging.WARNING))
        log.addHandler(log_stdouthandler)

        log_stderrhandler = logging.StreamHandler(sys.stderr)
        log_stderrhandler.setLevel(logging.ERROR)
        log.addHandler(log_stderrhandler)

    if verbose and verbose >= 2:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)


def load_module(category, name):
    module_name = "%s.%s" % (category, name)
    if __package__:
        module_name = "." + module_name
    try:
        module = importlib.import_module(module_name, package=__package__)
    except ImportError:
        logging.error("Module not found")
        raise

    return module


def get_initial_args_parser():
    config_dir = os.path.abspath(os.path.expanduser(
        os.environ.get("XDG_CONFIG_HOME")
        or os.environ.get("APPDATA")
        or os.path.join(os.environ["HOME"], ".config")
    ))

    initial_parser = argparse.ArgumentParser(
        description="dorkbot", add_help=False)
    initial_parser.add_argument("-c", "--config",
                                default=os.path.join(config_dir, "dorkbot", "dorkbot.ini"),
                                help="Configuration file")
    initial_parser.add_argument("-r", "--directory",
                                default=os.getcwd(),
                                help="Dorkbot directory (default location of db, tools, reports)")
    initial_parser.add_argument("--source", nargs="?", const=True, default=False,
                                help="Label associated with targets")
    initial_parser.add_argument("--show-defaults", action="store_true",
                                help="Show default values in help output")
    retrieval_options = initial_parser.add_argument_group("retrieval")
    retrieval_options.add_argument("--count", type=int, default=-1,
                                   help="number of targets to retrieve, or -1 for all")
    retrieval_options.add_argument("--random", action="store_true",
                                   help="retrieve targets in random order")
    initial_args, other_args = initial_parser.parse_known_args()

    return initial_args, other_args, initial_parser


def get_main_args_parser():
    initial_args, other_args, initial_parser = get_initial_args_parser()

    defaults = {
        "database": os.path.join(initial_args.directory, "dorkbot.db"),
    }

    if os.path.isfile(initial_args.config):
        config = configparser.ConfigParser()
        config.read(initial_args.config)
        try:
            config_items = config.items("dorkbot")
            defaults.update(dict(config_items))
        except KeyError:
            pass
        except configparser.NoSectionError as e:
            logging.debug(e)

    if initial_args.show_defaults:
        parser = argparse.ArgumentParser(parents=[initial_parser], add_help=False,
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    else:
        parser = argparse.ArgumentParser(parents=[initial_parser], add_help=False)

    parser.set_defaults(**defaults)
    parser.add_argument("-h", "--help", action="store_true",
                        help="Show program (or specified module) help")
    parser.add_argument("--log",
                        help="Path to log file")
    parser.add_argument("-v", "--verbose", action="count",
                        help="Enable verbose logging (can be used multiple times to increase verbosity)")
    parser.add_argument("-V", "--version", action="version",
                        version="%(prog)s " + __version__, help="Print version")

    database = parser.add_argument_group('database')
    database.add_argument("-d", "--database",
                          help="Database file/uri")
    database.add_argument("-u", "--prune", action="store_true",
                          help="Apply fingerprinting and blocklist without scanning")
    database.add_argument("--drop-tables", action="store_true",
                          help="Delete and recreate tables")
    database.add_argument("--retries", type=int, default=3,
                          help="Number of retries when an operation fails")
    database.add_argument("--retry-on", action="append", default=[],
                          help="Error strings that should result in a retry (can be used multiple times)")

    targets = parser.add_argument_group('targets')
    targets.add_argument("-l", "--list-targets", action="store_true",
                         help="List targets in database")
    targets.add_argument("-n", "--unscanned-only", action="store_true",
                         help="Only include unscanned targets")
    targets.add_argument("--list-sources", action="store_true",
                         help="List sources in database")
    targets.add_argument("--add-target", metavar="TARGET",
                         help="Add a url to the target database")
    targets.add_argument("--delete-target", metavar="TARGET",
                         help="Delete a url from the target database")
    targets.add_argument("--flush-targets", action="store_true",
                         help="Delete all targets")
    targets.add_argument("-e", "--delete-on-error", action="store_true",
                         help="Delete target if error encountered while processing it")

    indexing = parser.add_argument_group('indexing')
    indexing.add_argument("-i", "--indexer",
                          help="Indexer module to use")
    indexing.add_argument("-o", "--indexer-arg", action="append",
                          help="Pass an argument to the indexer module (can be used multiple times)")

    scanning = parser.add_argument_group('scanning')
    scanning.add_argument("-s", "--scanner",
                          help="Scanner module to use")
    scanning.add_argument("-p", "--scanner-arg", action="append",
                          help="Pass an argument to the scanner module (can be used multiple times)")
    scanning.add_argument("-t", "--test", action="store_true",
                          help="Fetch next scannable target but do not mark it scanned")
    scanning.add_argument("-x", "--reset-scanned", action="store_true",
                          help="Reset scanned status of all targets")

    fingerprints = parser.add_argument_group('fingerprints')
    fingerprints.add_argument("-g", "--generate-fingerprints", action="store_true",
                              help="Generate fingerprints for all targets")
    fingerprints.add_argument("-f", "--flush-fingerprints", action="store_true",
                              help="Delete all generated fingerprints")
    fingerprints.add_argument("--fingerprint-max", type=int, default=-1,
                              help="Maximum matches per fingerprint before deleting new matches")

    blocklist = parser.add_argument_group('blocklist')
    blocklist.add_argument("--list-blocklist", action="store_true",
                           help="List internal blocklist entries")
    blocklist.add_argument("--add-blocklist-item", metavar="ITEM",
                           help="Add an ip/host/regex pattern to the internal blocklist")
    blocklist.add_argument("--delete-blocklist-item", metavar="ITEM",
                           help="Delete an item from the internal blocklist")
    blocklist.add_argument("--flush-blocklist", action="store_true",
                           help="Delete all internal blocklist items")
    blocklist.add_argument("-b", "--external-blocklist", action="append",
                           help="Supplemental external blocklist file/db (can be used multiple times)")

    args = parser.parse_args(other_args, namespace=initial_args)
    return args, parser


def get_module_parser(module, parent_parser=None):
    initial_args, other_args, initial_parser = get_initial_args_parser()

    defaults = {}
    module_defaults = {}

    if os.path.isfile(initial_args.config):
        config = configparser.ConfigParser()
        config.read(initial_args.config)

        try:
            config_items = config.items("dorkbot")
            defaults.update(dict(config_items))
        except KeyError:
            pass
        except configparser.NoSectionError as e:
            logging.debug(e)

        try:
            module_config_items = config.items(module.__name__)
            module_defaults.update(dict(module_config_items))
        except KeyError:
            pass
        except configparser.NoSectionError as e:
            logging.debug(e)

    if parent_parser:
        initial_parser = parent_parser

    usage = "%(prog)s [args] -i/-s [module] -o/-p [module_arg[=value]] ..."
    epilog = "NOTE: module args are passed via -o/-p as key=value and do not themselves require hyphens"

    if initial_args.show_defaults:
        parser = argparse.ArgumentParser(parents=[initial_parser], usage=usage, epilog=epilog, add_help=False,
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    else:
        parser = argparse.ArgumentParser(parents=[initial_parser], usage=usage, epilog=epilog, add_help=False)

    parser.set_defaults(**defaults)
    module.populate_parser(initial_args, parser)
    parser.set_defaults(**module_defaults)

    return parser, other_args


def format_module_args(args_list):
    args = []

    if args_list:
        for arg in args_list:
            if arg.startswith("--"):
                args.append(arg)
            else:
                args.append("--" + arg)

    return args


def index(db, blocklists, indexer, args, indexer_args):
    indexer_name = indexer.__name__.split(".")[-1]
    logging.info(f"Indexing: {indexer_name} {vars(indexer_args) if args.verbose else ''}")
    setattr(indexer_args, "directory", args.directory)
    urls, module_source = indexer.run(indexer_args)
    if args.source:
        source = args.source
    else:
        source = module_source
    db.add_targets(urls, source=source, blocklists=blocklists)


def scan(db, blocklists, scanner, args, scanner_args):
    scanned = 0
    while scanned < args.count or args.count == -1:
        if args.test and scanned > 0:
            break
        url = db.get_next_target(args, blocklists=blocklists)
        if not url:
            break

        target = Target(url)
        logging.info(f"Scanning: {target.url} {vars(scanner_args) if args.verbose else ''}")
        start_time = generate_timestamp()
        results = scanner.run(scanner_args, target)
        end_time = generate_timestamp()
        scanned += 1

        if results is False:
            logging.error(f"Scan failed: {target.url}")
            continue

        report = generate_report(target.url, start_time, end_time, scanner_args.label, results)
        write_report(report, scanner_args, hash=target.get_hash())


if __name__ == "__main__":
    main()



def populate_general_options(module_group):
    module_group.add_argument("--retries", type=int, default=2,
                              help="number of times to retry fetching results on error")
    module_group.add_argument("--threads", type=int, default=1,
                              help="number of concurrent requests to wayback.org")



import os
import subprocess

if __package__:
    from .general import populate_general_options
else:
    from scanners.general import populate_general_options


def populate_parser(args, parser):
    scanner = __name__.split(".")[-1]
    module_group = parser.add_argument_group(__name__, f"Scans with the {scanner} command-line scanner")
    populate_general_options(args, module_group)
    module_group.add_argument("--path", default=os.path.join(args.directory, "tools", scanner, "bin"),
                              help="path to scanner binary")

    return parser


def run(args, target):
    scan_cmd = [os.path.abspath(os.path.join(os.sep, "bin", "echo"))]
    scan_cmd += ["pretending"]
    scan_cmd += ["to"]
    scan_cmd += ["scan", target.url]
    try:
        subprocess.run(scan_cmd, check=True)
    except subprocess.CalledProcessError:
        return False

    vulns = []
    for i in range(1, 3):
        vuln = {}
        vuln["vulnerability"] = "Example Vulnerability %d" % i
        vuln["url"] = target.url
        vuln["parameter"] = "foo%d" % i
        vuln["method"] = "post"
        vuln["poc"] = target.url + "';%20drop%20table%20users;--%20"
        vulns.append(vuln)

    return vulns



import os


def populate_general_options(args, module_group):
    module_group.add_argument("--args",
                              help="space-delimited list of additional arguments")
    module_group.add_argument("--report-dir", default=os.path.join(args.directory, "reports"),
                              help="directory to save report file")
    module_group.add_argument("--report-filename",
                              help="filename to save vulnerability report as")
    module_group.add_argument("--report-append",
                              help="append to report file if it exists", action="store_true")
    module_group.add_argument("--report-indent",
                              help="indent level for vulnerability report json")
    module_group.add_argument("--label", default="",
                              help="friendly name field to include in vulnerability report")



import json
import logging
import time
from urllib.error import HTTPError
from urllib.parse import urlencode, urlparse
from urllib.request import Request, urlopen


def populate_parser(args, parser):
    module_group = parser.add_argument_group(__name__, "Searches bing.com")
    module_group.add_argument("--key", required=True,
                              help="API key")
    module_group.add_argument("--query", required=True,
                              help="search query")


def run(args):
    source = __name__.split(".")[-1]
    results = get_results(args.key, args.query)
    for result in results:
        logging.debug(result)
    logging.info("Fetched %d results", len(results))
    return results, source


def get_results(key, query):
    data = {"q": query,
            "count": 50,
            "offset": 0}

    results = []
    while data["offset"] < 1000:
        items = issue_request(data, key)
        data["offset"] += data["count"]
        if not items:
            break
        results.extend(items)

    return results


def issue_request(data, key):
    url = "https://api.bing.microsoft.com/v7.0/search?" + urlencode(data)
    while True:
        try:
            r = Request(url)
            r.add_header("Ocp-Apim-Subscription-Key", key)
            response_str = urlopen(r)
            response_str = response_str.read().decode("utf-8")
            response = json.loads(response_str)
            break
        except HTTPError as e:
            response_str = e.read().decode("utf-8")
            response = json.loads(response_str)
            if e.code == 429:
                time.sleep(0.5)

    if "webPages" not in response or response["webPages"]["totalEstimatedMatches"] < data["offset"]:
        return []

    return [urlparse(item["url"].strip()).geturl() for item in response["webPages"]["value"]]
